import logging

from mllaunchpad import ModelInterface, ModelMakerInterface


logger = logging.getLogger(__name__)

# Train this example from the command line:
# python -m mllaunchpad -c addition_cfg.yml train
#
# Start REST API:
# python -m mllaunchpad -c addition_cfg.yml api
#
# Example API call:
# http://127.0.0.1:5000/add/v0/sum?x1=3&x2=2


class MyAdditionExampleModelMaker(ModelMakerInterface):
    """Creates a model which adds (this can't really be called training)
    """

    def create_trained_model(self, model_conf, data_sources, data_sinks, old_model=None):

        # We can get info from the model's config dictionary if necessary
        my_name = model_conf["name"]
        logger.info(
            "%s is doing lots of important preparation for addition...", my_name
        )

        # Imagine we are preparing data and training a sklearn or some other model here
        def add(x1, x2):
            return x1 + x2
        my_lame_predictor = add

        # Usually, we put one or several trained regressors/classifiers into the
        # model, but in this example we just use our function.
        model = {"lame_pred": my_lame_predictor}
        # In addition to regressors/classifiers/etc., you can also use this to
        # tell prediction about e.g. calculated parameters, thresholds, etc.
        # that have been calculated during the training process.
        # For parameters that are stable from one training to the next, please
        # use the config's model: prediction_config/training_config instead.

        # First return value: model, second value: metrics dict
        return model

    def test_trained_model(self, model_conf, data_sources, data_sinks, model):

        # Okay, this is a stupid stand-in for preparing test data (usually you
        # have a specific data source for that)
        x1_test = 3
        x2_test = 7
        y_test = 10

        # Prediction happens here. Usually (though you *can*), you will not call
        # your own predict method, but get the inner model (classifier/regressor/...)
        # and use it directly.
        lame_predictor = model["lame_pred"]
        y = lame_predictor(x1_test, x2_test)

        # Calculate some metrics
        acc = 1 if y == y_test else 0  # haha, I know

        # Adhere to consistent lower-case naming (accuracy, confusion_matrix, f1, etc.),
        # don't invent new names for common test metrics that will possibly be read
        # automatically in the model's life cycle management. You can add new
        # metrics if you want, just don't invent abbreviations like 'acc' or such.
        metrics = {"accuracy": acc, "precision": "whatever"}

        return metrics


class MyAdditionExampleModel(ModelInterface):
    """Simplest possible 'Data Science Model' example, without using data sources.
    """

    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):

        # We can optionally get info from the model's config dictionary
        my_name = model_conf["name"]
        logger.info("Hey, look at me, %s -- I'm adding two numbers!", my_name)

        # Get the parameters/arguments from the API call
        x1 = args_dict["x1"]
        x2 = args_dict["x2"]

        # Get the predictor from the dict that we stuffed into this model when training
        my_lame_predictor = model["lame_pred"]

        # Our sophisticated model does its magic here
        y = my_lame_predictor(x1, x2)

        # We usually return a dict with prediction results
        return {"my_addition_result": y}



import logging

# Third-party imports
from mllaunchpad import ModelInterface, ModelMakerInterface, order_columns
import pandas as pd
from sklearn import tree
from sklearn.metrics import accuracy_score, confusion_matrix


logger = logging.getLogger(__name__)

# Train this example from the command line:
# python -m mllaunchpad -c tree_cfg.yml train
#
# Start REST API:
# python -m mllaunchpad -c tree_cfg.yml api
#
# Example API call:
# http://127.0.0.1:5000/iris/v0/varieties?sepal.length=4.9&sepal.width=2.4&petal.length=3.3&petal.width=1
#
# Example with URI Param (Resource ID):
# http://127.0.0.1:5000/iris/v0/varieties/12?hallo=metric
#
# Example to trigger batch prediction (not really the idea of an API...):
# http://127.0.0.1:5000/iris/v0/varieties


class MyExampleModelMaker(ModelMakerInterface):
    """Creates a model
    """

    def create_trained_model(
        self, model_conf, data_sources, data_sinks, old_model=None
    ):

        df = data_sources["petals"].get_dataframe()
        X = df.drop("variety", axis=1)
        y = df["variety"]

        X_ordered = order_columns(X)

        my_tree = tree.DecisionTreeClassifier()
        my_tree.fit(X_ordered, y)

        return my_tree

    def test_trained_model(self, model_conf, data_sources, data_sinks, model):
        df = data_sources["petals_test"].get_dataframe()
        X_test = order_columns(df.drop("variety", axis=1))
        y_test = df["variety"]

        my_tree = model

        y_predict = my_tree.predict(X_test)

        acc = accuracy_score(y_test, y_predict)
        conf = confusion_matrix(y_test, y_predict).tolist()

        metrics = {"accuracy": acc, "confusion_matrix": conf}

        return metrics


class MyExampleModel(ModelInterface):
    """Uses the created Data Science Model
    """

    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):

        if "test_key" in args_dict:
            # URI param example (an uri param is part in the args_dict just like any other input)
            key = args_dict["test_key"]
            logger.info(
                "Got the uri parameter (ID) %s. Looking up input data and predicting...",
                key,
            )
            df = data_sources["batch_input"].get_dataframe()
            df["myid"] = df["myid"].apply(str)
            X = order_columns(df.loc[df["myid"] == key])
            my_tree = model
            y = my_tree.predict(X.drop("myid", axis=1))[0]
            return {"iris_variety": y}
        elif "sepal.length" not in args_dict or args_dict["sepal.length"] is None:
            # Batch prediction example
            logger.info("Doing batch prediction")
            X = order_columns(data_sources["batch_input"].get_dataframe())
            my_tree = model
            y = my_tree.predict(X.drop("myid", axis=1))
            X["pred"] = y
            data_sinks["predictions"].put_dataframe(X)
            return {"status": "ok"}

        # "Normal" prediction example
        logger.info('Doing "normal" prediction')
        X = order_columns(pd.DataFrame(args_dict, index=[0]))

        my_tree = model
        y = my_tree.predict(X)[0]

        return {"iris_variety": y}



import logging

from mllaunchpad import ModelInterface, ModelMakerInterface
import rpy2
import rpy2.rinterface as ri
import rpy2.robjects as ro
import rpy2.robjects.packages as rp
from rpy2.robjects import pandas2ri
# from rpy2.robjects import numpy2ri


logger = logging.getLogger(__name__)
# numpy2ri.activate()  # vector=rpyn.ri2numpy(vector_R)
pandas2ri.activate()  # If wishing to convert explicitly for any reason, the functions are pandas2ri.py2rpy() and pandas2ri.rpy2py (rpy2 >= 3.0.0) (rpy2 2.9.4: pandas2ri.py2ri() and pandas2ri.ri2py(), which earlier were pandas2ri.pandas2ri() and pandas2ri.ri2pandas())).

if rpy2.__version__[0] == "2":  # tested with rpy2 == 2.9.4
    rpy2_v2 = True
    py2rpy = pandas2ri.py2ri
    rpy2py = pandas2ri.ri2py
else:  # tested with rpy2 == 3.0.4
    rpy2_v2 = False
    py2rpy = ro.conversion.py2rpy
    rpy2py = ro.conversion.rpy2py


def load_dependencies(model_conf):
    if not hasattr(ro.r, "loaded_r_model_dependencies"):
        utils = rp.importr("utils")
        package_names = model_conf["r_dependencies"]
        for pkg in package_names:
            loaded = ro.r.require(pkg)
            if not loaded[0]:
                logger.info("R package %s is not installed. Installing...", pkg)
                utils.install_packages(pkg)
                logger.info("Loading freshly installed R package %s...", pkg)
                ro.r.library(pkg)
            logger.info("Loaded R dependency package %s", pkg)
        ro.globalenv["loaded_r_model_dependencies"] = 1


def source_r_file(model_conf):
    if not hasattr(ro.r, "create_trained_model") and not hasattr(
        ro.r, "predict_with_trained_model"
    ):
        r_filename = model_conf["r_file"]
        logger.info("Loading R source file: %s...", r_filename)
        ro.r.source(r_filename)


def dict_to_r_list(d):
    if type(d) is not dict:
        return d
    else:
        return ro.r.list(**{key: dict_to_r_list(val) for key, val in d.items()})


def r_list_to_dict(d):
    if type(d) is ro.FactorVector:
        return [l for l in d.iter_labels()]
    elif type(d) is not ro.ListVector:
        return d
    else:
        if rpy2_v2:
            return {
                key: r_list_to_dict(d.rx2(key))[0] for key in d.names
            }  # rpy2 == 2.9.4
        else:
            return {key: r_list_to_dict(val[0]) for key, val in d.items()}


#  (rpy2 >= 3.0.x) Workaround to keep functions from being garbage-collected
aref = None
aref2 = None


def prepare_r_data_sources_and_sinks(data_sources, data_sinks, kind):
    """Makes data source/sink access functions available in R environment
    """
    if not hasattr(ro.r, "created_data_sources_and_sinks_" + kind):
        logger.debug(
            "Creating R get_dataframe and put_dataframe functions for %s", kind
        )

        @ri.rternalize
        def get_dataframe(name_vec, arg_dict=None):
            name = name_vec[0]
            logger.debug(
                "Calling python DS %s's get_dataframe and returning R dataframe", name
            )
            pd_df = data_sources[name].get_dataframe(arg_dict)
            return py2rpy(pd_df)

        @ri.rternalize
        def put_dataframe(name_vec, r_df, arg_dict=None):
            name = name_vec[0]
            logger.debug(
                "Calling python DS %s's put_dataframe with ri2py converted pandas dataframe",
                name,
            )
            df = rpy2py(r_df)
            data_sinks[name].put_dataframe(df, arg_dict)

        ro.globalenv["get_dataframe"] = get_dataframe
        ro.globalenv["put_dataframe"] = put_dataframe

        # (rpy2 >= 3.0.x) Workaround to keep functions from being garbage-collected
        global aref, aref2
        aref = get_dataframe
        aref2 = put_dataframe

        ro.globalenv["created_data_sources_and_sinks_" + kind] = 1
    else:
        logger.debug(
            "Using existing R get_dataframe and put_dataframe functions for %s", kind
        )


class MyExampleModelMaker(ModelMakerInterface):
    """Creates an R-based model
    """

    def create_trained_model(self, model_conf, data_sources, data_sinks, old_model=None):
        load_dependencies(model_conf)
        source_r_file(model_conf)
        prepare_r_data_sources_and_sinks(data_sources, data_sinks, "train")

        r_model_conf = dict_to_r_list(model_conf)
        r_old_model = old_model if old_model is not None else ro.r("NULL")
        logger.debug("Calling R function test_trained_model")
        model = ro.r.create_trained_model(r_model_conf, r_old_model)
        logger.debug("Returned from R function create_trained_model")

        return model

    def test_trained_model(self, model_conf, data_sources, data_sinks, model):
        load_dependencies(model_conf)
        source_r_file(model_conf)
        prepare_r_data_sources_and_sinks(data_sources, data_sinks, "test")

        r_model_conf = dict_to_r_list(model_conf)
        r_model = model
        logger.debug("Calling R function test_trained_model")
        r_metrics = ro.r.test_trained_model(r_model_conf, r_model)
        logger.debug("Returned from R function test_trained_model")
        metrics = r_list_to_dict(r_metrics)

        return metrics


class MyExampleModel(ModelInterface):
    """Uses the created Data Science Model
    """

    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):
        load_dependencies(model_conf)
        source_r_file(model_conf)
        prepare_r_data_sources_and_sinks(data_sources, data_sinks, "predict")

        r_args_list = dict_to_r_list(dict(args_dict))
        r_model_conf = dict_to_r_list(model_conf)
        r_model = model
        logger.debug("Calling R function predict_with_model")
        r_output = ro.r.predict_with_model(r_model_conf, r_args_list, r_model)
        logger.debug("Returned from R function predict_with_model")

        output = r_list_to_dict(r_output)

        return output



import logging

from mllaunchpad.resource import DataSource
import pandas as pd


logger = logging.getLogger(__name__)


class BogusDataSource(DataSource):
    """DataSource for creating nonsense
    """

    serves = ["bogus"]

    # def __init__(self, identifier, datasource_config):
    #     super().__init__(identifier, datasource_config)

    def get_dataframe(self, params=None, buffer=False):
        """Get some pandas dataframe.

        Params:
            argsDict: optional, currently not implemented
            buffer: optional, currently not implemented

        Returns:
            DataFrame object, possibly cached according to expires-config
        """
        if buffer:
            raise NotImplementedError("Buffered reading not supported yet")

        kw_options = self.options

        return pd.DataFrame({"a": [3, 4, 5], "b": [6, 7, 8]}, **kw_options)

    def get_raw(self, params=None, buffer=False):
        """Not implemented.

        Params:
            argsDict: optional, currently not implemented
            buffer: optional, currently not implemented

        Returns:
            Nothing, throws NotImplementedError
        """
        raise NotImplementedError("Raw bogus not supported yet")



# Stdlib imports
import logging
import logging.config
import os
import warnings

# Third-party imports
import yaml


LOG_CONF_FILENAME_DEFAULT = "./LAUNCHPAD_LOG.yml"
LOG_CONF_FILENAME_ENV = os.environ.get(
    "LAUNCHPAD_LOG", LOG_CONF_FILENAME_DEFAULT
)


def init_logging(filename=LOG_CONF_FILENAME_ENV, verbose=False):
    """Only called from wsgi or cli module (mllaunchpad-as-an-app).
    It's important to not change logging/warning config from the library-only
    code.
    """
    # Ignore all deprecation warnings:
    warnings.filterwarnings(action="ignore", category=DeprecationWarning)
    # Except from mllaunchpad itself:
    warnings.filterwarnings(
        action="default", category=DeprecationWarning, module="mllaunchpad.*"
    )
    try:
        loaded_logging_config = yaml.safe_load(open(filename))
        logging.config.dictConfig(loaded_logging_config)
    except FileNotFoundError:
        loaded_logging_config = None
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(name)s: %(message)s",
            level=logging.INFO,
        )

    logging.captureWarnings(True)
    new_logger = logging.getLogger(__name__)
    if verbose:
        new_logger.setLevel(logging.DEBUG)

    if not loaded_logging_config:
        new_logger.warning(
            "Logging filename environment variable LAUNCHPAD_LOG not set, "
            "and ./LAUNCHPAD_LOG.yml not found: using default logging "
            "configuration"
        )
    elif filename == LOG_CONF_FILENAME_DEFAULT:
        new_logger.warning(
            "Logging filename environment variable LAUNCHPAD_LOG not set, "
            "using default logging configuration file ./LAUNCHPAD_LOG.yml"
        )

    return new_logger



import logging
import os

import pandas as pd

from mllaunchpad.resource import DataSource


logger = logging.getLogger(__name__)

try:
    from impala.dbapi import connect
except ModuleNotFoundError:
    logger.warning("Please install the impyla package to be able to use ImpalaDataSource.")


class ImpalaDataSource(DataSource):
    """DataSource for fetching data from Impala queries.

    EXPERIMENTAL

    Creates a short-lived connection on every non-cached call to `get_dataframe`.

    Configuration::

        dbms:
          # ... (other connections)
          my_connection:  # NOTE: You can use the same connection for several datasources and datasinks
            type: impala
            # NOTE: The options below are examples for accessing a kerberized cluster.
            #       You can specify other options here, and they will be used for creating
            #       the connection. For a full list, refer to the documentation of
            #       `impyla.dbapi.connect`: https://pydoc.net/impyla/0.14.0/impala.dbapi/
            # NOTE 2: For user and password parameters, you can provide the parameter with
            #         the suffix "_var". For example, if you configure `ldap_password_var: MY_PW`,
            #         ML Launchpad will get the value of the environment variable `MY_PW` and pass
            #         it to `impyla.dbapi.connect`s `ldap_password` parameter. This is to avoid
            #         having to add passwords to the configuration file.
            host: host.example.com
            port: 21050
            database: my_db
            kerberos_service_name: impala
            auth_mechanism: GSSAPI
            use_ssl: true
        # ...
        datasources:
          # ... (other datasources)
          my_datasource:
            type: dbms.my_connection
            query: SELECT * FROM my_table where id = :id  # fill `:params` by calling `get_dataframe` with a `dict`
            expires: 0     # generic parameter, see documentation on DataSources
            tags: [train]  # generic parameter, see documentation on DataSources and DataSinks
            options: {}    # (optional) any other kwargs to pass to `pd.read_sql`

    Note: In order to access a kerberized cluster, you need the correct packages.
    We found the following packages to work reliably::

        $ pip install pykerberos thrift_sasl impyla pandas
        (for Windows, use winkerberos instead of pykerberos)
    """
    serves = ['dbms.impala']

    def __init__(self, identifier, datasink_config, dbms_config):
        super().__init__(identifier, datasink_config)

        # Fill "_var"-suffixed configuration items from environment variables
        new_dbms_config = {}
        key: str
        for key, value in dbms_config.items():
            if key.endswith("_var"):
                new_value = os.environ.get(value)
                if new_value is not None:
                    new_key = key[:-4]
                    new_dbms_config[new_key] = new_value
                    logger.debug("Replaced Impala connection parameter '%s' specifying environment"
                                 "variable '%s' with parameter '%s'", key, value, new_key)
                else:
                    logger.warning("Environment variable '%s' not set (from config key '%s')", value, key)
                    new_dbms_config[key] = value
            else:
                new_dbms_config[key] = value

        self.dbms_config = new_dbms_config

    def get_dataframe(self, params=None, buffer=False):
        """Get data as a pandas dataframe.

        Example::

            data_sources["my_datasource"].get_dataframe({"id": 387})

        :param params: Query parameters to fill in query (e.g. `:id` with value 387)
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :return: DataFrame object, possibly cached according to expires-config
        """
        if buffer:
            raise NotImplementedError('Buffered reading not supported yet')

        # Open connection
        logger.info(
            "Establishing Impala database connection for datasource {}...".format(
                self.id
            )
        )
        conn_args = {k: v for k, v in self.dbms_config.items() if k not in ["type"]}
        with connect(**conn_args) as conn:
            # Fetch query
            query = self.config["query"]
            params = params or {}
            kw_options = self.config.get("options", {})
            logger.debug(
                "Fetching query {} with params {} and options {}...".format(
                    query, params, kw_options
                )
            )
            df = pd.read_sql(
                query, con=conn, params=params, **kw_options
            )

        return df

    def get_raw(self, params=None, buffer=False):
        """Not implemented.

        :param params: Query parameters to fill in query (e.g. `:id` with value 387)
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :raises NotImplementedError:

        :return: Nothing, always raises NotImplementedError
        """
        raise NotImplementedError(
            "ImpalaDataSource currently does not not support raw format/blobs. "
            'Use method "get_dataframe" for dataframes'
        )









# Stdlib imports
import abc
import getpass
import glob
import json
import logging
import os
import shutil
import sys
from collections import OrderedDict
from datetime import datetime
from time import time
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

# Third-party imports
# We are only unpickling files which are completely under the
# control of the model developer, not influenced by end user data.
import dill as pickle  # nosec
import numpy as np
import pandas as pd


DS = TypeVar("DS", "DataSource", "DataSink")
Raw = Union[str, bytes]


logger = logging.getLogger(__name__)

SUPPORTED_FILE_TYPES = ["csv", "euro_csv", "text_file", "binary_file"]
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
DATE_FORMAT_FILES = "%Y-%m-%d_%H-%M-%S"


class ModelStore:
    """Deals with persisting, loading, updating metrics metadata of models.
    Abstracts away how and where the model is kept.

    TODO: Smarter querying like 'get me the model with the currently (next)
    best metrics which serves a particular API.'
    """

    def __init__(self, config):
        """Get a model store based on the config settings

        Params:
            config: configuration dict
        """
        self.location = config["model_store"]["location"]
        if not os.path.exists(self.location):
            os.makedirs(self.location)

    def _get_model_base_name(self, model_conf):
        return os.path.join(
            self.location,
            "{}_{}".format(model_conf["name"], model_conf["version"]),
        )

    @staticmethod
    def _load_metadata(base_name):
        metadata_name = base_name + ".json"
        with open(metadata_name, "r") as f:
            meta = json.load(f)

        return meta

    @staticmethod
    def _dump_metadata(base_name, raw_metadata):
        metadata_name = base_name + ".json"
        metadata = to_plain_python_obj(raw_metadata)
        try:
            with open(metadata_name, "w") as f:
                json.dump(metadata, f, indent=2)
        except TypeError as e:
            os.remove(metadata_name)
            raise e

    def _backup_old_model(self, base_name):
        backup_dir = os.path.join(self.location, "previous")
        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)
        infix = datetime.now().strftime(DATE_FORMAT_FILES)
        for file in glob.glob(base_name + "*"):
            fn_ext = os.path.basename(file)
            fn, ext = os.path.splitext(fn_ext)
            new_file_name = "{}_{}{}".format(fn, infix, ext)
            logger.debug(
                "Backing up previous model file {} as {}".format(
                    fn_ext, new_file_name
                )
            )
            shutil.copy(file, os.path.join(backup_dir, new_file_name))

    def dump_trained_model(self, complete_conf, model, metrics):
        """Save a model object in the model store. Some metadata will also
        be saved along the model, including the metrics which is the second parameter.

        Params:
            model_conf:  the config dict of our model
            model:       the model object to store
            metrics:     metrics dictionary

        Returns:
            Nothing
        """
        model_conf = complete_conf["model"]
        base_name = self._get_model_base_name(model_conf)

        # Check if exists and backup if it does
        self._backup_old_model(base_name)

        # Save model itself
        pkl_name = base_name + ".pkl"
        with open(pkl_name, "wb") as f:
            pickle.dump(model, f)

        # Save metadata
        meta = {
            "name": model_conf["name"],
            "version": model_conf["version"],
            "created": datetime.now().strftime(DATE_FORMAT),
            "created_by": getpass.getuser(),
            "metrics": metrics,
            "metrics_history": {datetime.now().strftime(DATE_FORMAT): metrics},
            "config_snapshot": model_conf,
        }
        if "api" in complete_conf:  # API is optional
            meta["api_name"] = complete_conf["api"]["name"]

        self._dump_metadata(base_name, meta)

    def load_trained_model(self, model_conf):
        """Load a model object from the model store. Some metadata will also
        be loaded along the model.

        Params:
            model_conf:  the config dict of our model

        Returns:
            Tuple of model object and metadata dictionary
        """
        base_name = self._get_model_base_name(model_conf)

        if "." not in sys.path:
            sys.path.append(".")

        pkl_name = base_name + ".pkl"
        with open(pkl_name, "rb") as f:
            # We are only unpickling files which are completely under the
            # control of the model developer, not influenced by end user data.
            model = pickle.load(f)  # nosec

        meta = self._load_metadata(base_name)

        return model, meta

    def update_model_metrics(self, model_conf, metrics):
        """Update the test metrics for a previously stored model
        """
        base_name = self._get_model_base_name(model_conf)
        meta = self._load_metadata(base_name)
        meta["metrics"] = metrics
        meta["metrics_history"][datetime.now().strftime(DATE_FORMAT)] = metrics
        self._dump_metadata(base_name, meta)


def _tags_match(tags, other_tags) -> bool:
    tags = tags or []
    if isinstance(tags, str):
        tags = [tags]

    other_tags = other_tags or []
    if isinstance(other_tags, str):
        other_tags = [other_tags]

    tags_required = bool(tags)
    tags_provided = bool(other_tags)
    tags_matching = bool(set(tags) & set(other_tags))

    return not tags_required or not tags_provided or tags_matching


def _get_all_classes(config, the_type: Type[DS]) -> Dict[str, Type[DS]]:
    modules = [__name__]  # find built_in types using same mechanism
    if "plugins" in config:
        logger.info("Loading %s plugins", the_type)
        # Append plugins so they can replace builtin types
        modules += config["plugins"]

    ds_cls: Dict[str, Type[DS]] = {}
    for module in modules:
        __import__(module)
        # Handle one import after another so plugins can replace builtin types
        imported_classes = [
            cls
            for cls in the_type.__subclasses__()
            if cls.__module__ == module
        ]
        for cls in imported_classes:
            if hasattr(cls, "serves") and hasattr(cls.serves, "__iter__"):
                for k in cls.serves:
                    if k in ds_cls:
                        logger.warning(
                            f"Plugin class {cls} shadows {ds_cls[k]} which also serves {k}"
                        )
                    ds_cls[k] = cls
                logger.debug(
                    "Loaded %s.%s, serving %s}",
                    module,
                    cls.__name__,
                    cls.serves,
                )
            else:
                logger.warning(f'Class {cls} has no list attribute "serves"')
    return ds_cls


def _create_data_sources_or_sinks(
    config: Dict, the_type: Type[DS], tags: Iterable[str] = []
) -> Dict[str, DS]:
    # Implementation note: no generator used because we want to fail early
    if the_type == DataSource:
        what = "datasource"
        config_key = "datasources"
    elif the_type == DataSink:
        what = "datasink"
        config_key = "datasinks"
    else:
        raise TypeError(
            "`the_type` parameter must be `DataSource` or `DataSink`"
        )
    ds_objects: Dict[str, DS] = {}

    ds_cls: Dict[str, Type[DS]] = _get_all_classes(config, the_type)
    logger.debug("ds_cls=%s", ds_cls)

    if config_key not in config or not isinstance(config[config_key], dict):
        logger.info("No %s defined in configuration", config_key)
        return ds_objects

    for ds_id in config[config_key]:
        ds_config = config[config_key][ds_id]

        if not _tags_match(tags, ds_config.get("tags")):
            continue

        ds_types = ds_config["type"].split(".")
        main_type = ds_types[0]
        sub_type = ds_types[1] if len(ds_types) >= 2 else None
        ds_subtype_config = config[main_type][sub_type] if sub_type else None

        service_need = main_type + (
            "." + ds_subtype_config["type"] if ds_subtype_config else ""
        )

        if service_need not in ds_cls:
            raise ValueError(
                f"No {what} class for {service_need} available. Check the configuration for typos in the {what} type or add a suitable plugin."
            )

        logger.debug(
            "Initializing %s %s of type %s...", what, ds_id, ds_config["type"]
        )
        if ds_subtype_config is None:
            ds_objects[ds_id] = ds_cls[service_need](ds_id, ds_config)
        else:
            ds_objects[ds_id] = ds_cls[service_need](
                ds_id, ds_config, ds_subtype_config
            )

        logger.debug("%s %s initialized", what.capitalize(), ds_id)

    # typing.cast(Dict[str, DS], ds_objects)
    return ds_objects


def create_data_sources_and_sinks(
    config: Dict, tags: Iterable[str] = []
) -> Tuple[Dict[str, "DataSource"], Dict[str, "DataSink"]]:
    """Creates the data sources as defined in the configuration dict.
    Filters them by tag.

    Params:
        config: configuration dictionary
        tags:   optionally filter for only matching datasources no value(s) = match all datasources

    Returns:
        dict with keys=datasource names, values=initialized DataSource objects
    """

    # mypy is unfortunately only accepting instantiable classes here (no ABCs),
    # which is a know issue: https://github.com/python/mypy/issues/5374
    # Ignoring check for now, hoping for a solution/workaround soon.
    sources: Dict[str, DataSource] = _create_data_sources_or_sinks(
        config, the_type=DataSource, tags=tags  # type: ignore
    )
    sinks: Dict[str, DataSink] = _create_data_sources_or_sinks(
        config, the_type=DataSink, tags=tags  # type: ignore
    )

    return sources, sinks


class CacheDict(OrderedDict):
    def __init__(self, *args, **kwds):
        self.maxsize = kwds.pop("maxsize", None)
        OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key, value):
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.maxsize is not None:
            while len(self) > self.maxsize:
                self.popitem(last=False)

    def __hash__(self):
        return hash(frozenset(self.items()))


class CachedDataSource(type):
    """Metaclass to Auto-apply decorators "@cached" to data getters.
    https://stackoverflow.com/questions/10067262/automatically-decorating-every-instance-method-in-a-class
    """

    def __new__(mcs, name, bases, dct):
        for attr, func in dct.items():
            if attr.startswith("get_"):
                dct[attr] = CachedDataSource.cached(func)
        return type.__new__(mcs, name, bases, dct)

    @classmethod
    def cached(mcs, func):
        """This decorator is automatically applied to `get_dataframe` and `get_raw` methods to enable caching."""

        def wrapper(self, params: Dict = None, buffer: bool = False):
            key = (func.__name__, json.dumps(params, sort_keys=True), buffer)
            item = self._get_cached(key)
            if item is not None:
                return item
            else:
                result = func(self, params, buffer)
                self._to_cache(key, result)
                return result

        wrapper.__doc__ = func.__doc__
        return wrapper


class DataSource(metaclass=CachedDataSource):
    """Interface, used by the Data Scientist's model to get its data from.
    Concrete DataSources (for files, data bases, etc.) need to inherit from this class.
    """

    serves: List[str] = []

    def __init__(
        self,
        identifier: str,
        datasource_config: Dict,
        sub_config: Optional[Dict] = None,  # used in DBMS subclasses
    ):
        """Please call super().__init(...) when overwriting this method
        """
        self.id = identifier
        self.config = datasource_config
        self.options = self.config.get("options", {})

        self.expires = self.config.get("expires", 0)

        maxsize = self.config.get("cache_size", 32)
        self._cache = CacheDict(maxsize=maxsize)

    @abc.abstractmethod
    def get_dataframe(
        self, params: Dict = None, buffer: bool = False
    ) -> pd.DataFrame:
        ...

    @abc.abstractmethod
    def get_raw(self, params: Dict = None, buffer: bool = False) -> Raw:
        ...

    def _get_cached(self, key) -> Any:
        if self.expires == -1 or self.expires > 0:
            item, time_stamp = self._cache.get(key, (None, 0))
            if item is not None and (
                self.expires == -1 or time() <= time_stamp + self.expires
            ):
                logger.debug(
                    "Returning cached item for datasource %s", self.id
                )
                return item
        return None  # either immediately expires (0) or has expired in meantime (>0)

    def _to_cache(self, key, item) -> None:
        if self.expires != 0:
            self._cache[key] = (item, time())

    def __del__(self):
        """Overwrite to clean up any resources (connections, temp files, etc.).
        """
        ...


def get_user_pw(user_var: str, password_var: str) -> Tuple[str, Optional[str]]:
    user = os.environ.get(user_var)
    pw = os.environ.get(password_var)
    if user is None:
        raise ValueError(
            "User name environment variable {} not set".format(user_var)
        )
    if pw is None:
        logger.warning(
            "Password environment variable %s not set", password_var
        )
    return user, pw


def _get_oracle_connection(dbms_config: Dict):
    import cx_Oracle  # Importing here avoids environment-specific dependencies

    user, pw = get_user_pw(
        dbms_config["user_var"], dbms_config["password_var"]
    )
    dsn_tns = cx_Oracle.makedsn(
        dbms_config["host"],
        dbms_config["port"],
        service_name=dbms_config["service_name"],
    )
    logger.debug("Oracle connection string: %s", dsn_tns)

    kw_options = dbms_config.get("options", {})
    connection = cx_Oracle.connect(user, pw, dsn_tns, **kw_options)

    return connection


class OracleDataSource(DataSource):
    """DataSource for Oracle database connections.

    Creates a long-living connection on initialization.

    Configuration example::

        dbms:
          # ... (other connections)
          my_connection:  # NOTE: You can use the same connection for several datasources and datasinks
            type: oracle
            host: host.example.com
            port: 1251
            user_var: MY_USER_ENV_VAR
            password_var: MY_PW_ENV_VAR  # optional
            service_name: servicename.example.com
            options: {}  # used as **kwargs when initializing the DB connection
        # ...
        datasources:
          # ... (other datasources)
          my_datasource:
            type: dbms.my_connection
            query: SELECT * FROM somewhere.my_table where id = :id  # fill `:params` by calling `get_dataframe` with a `dict`
            expires: 0    # generic parameter, see documentation on DataSources
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when fetching the query using `pandas.read_sql`
    """

    serves = ["dbms.oracle"]

    def __init__(
        self, identifier: str, datasource_config: Dict, dbms_config: Dict
    ):
        super().__init__(identifier, datasource_config)

        self.dbms_config = dbms_config

        logger.info(
            "Establishing Oracle database connection for datasource {}...".format(
                self.id
            )
        )
        self.connection = _get_oracle_connection(dbms_config)

    def get_dataframe(
        self, params: Dict = None, buffer: bool = False
    ) -> pd.DataFrame:
        """Get the data as pandas dataframe.

        Example::

            data_sources["my_datasource"].get_dataframe({"id": 387})

        :param params: Query parameters to fill in query (e.g. replace query's `:id` parameter with value `387`)
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :return: DataFrame object, possibly cached according to config value of `expires:`
        """
        if buffer:
            raise NotImplementedError("Buffered reading not supported yet")

        # TODO: maybe want to open/close connection on every method call (shouldn't happen often)
        query = self.config["query"]
        params = params or {}
        kw_options = self.options

        logger.debug(
            "Fetching query {} with params {} and options {}...".format(
                query, params, kw_options
            )
        )
        df = pd.read_sql(
            query, con=self.connection, params=params, **kw_options
        )
        df.fillna(np.nan, inplace=True)

        return df

    def get_raw(self, params: Dict = None, buffer: bool = False) -> Raw:
        """Not implemented.

        :raises NotImplementedError: Raw/blob format currently not supported.
        """
        raise NotImplementedError(
            "OracleDataSource currently does not not support raw format/blobs. "
            'Use method "get_dataframe" for dataframes'
        )

    def __del__(self):
        if hasattr(self, "connection"):
            self.connection.close()


class FileDataSource(DataSource):
    """DataSource for fetching data from files.

    See :attr:`serves` for the available types.

    Configuration example::

        datasources:
          # ... (other datasources)
          my_datasource:
            type: euro_csv  # `euro_csv` changes separators to ";" and decimals to "," w.r.t. `csv`
            path: /some/file.csv  # Can be URL, uses `pandas.read_csv` internally
            expires: 0    # generic parameter, see documentation on DataSources
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when fetching the data using `pandas.read_csv`
          my_raw_datasource:
            type: text_file  # raw files can also be of type `binary_file`
            path: /some/file.txt  # Can be URL
            expires: 0    # generic parameter, see documentation on DataSources
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when fetching the data using `fh.read`
    """

    serves = SUPPORTED_FILE_TYPES

    def __init__(self, identifier: str, datasource_config: Dict):
        super().__init__(identifier, datasource_config)

        ds_type = datasource_config["type"]
        if ds_type not in SUPPORTED_FILE_TYPES:
            raise TypeError(
                "{} is not a datasource file type (in datasource {}).".format(
                    repr(ds_type), repr(identifier)
                )
            )

        self.type = ds_type
        self.path = datasource_config["path"]

    def get_dataframe(
        self, params: Dict = None, buffer: bool = False
    ) -> pd.DataFrame:
        """Get data as a pandas dataframe.

        Example::

            data_sources["my_datasource"].get_dataframe()

        :param params: Currently not implemented
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :return: DataFrame object, possibly cached according to config value of `expires:`
        """
        if buffer:
            raise NotImplementedError("Buffered reading not supported yet")

        kw_options = self.options

        logger.debug(
            "Loading type {} file {} with options {}...".format(
                self.type, self.path, kw_options
            )
        )
        if self.type == "csv":
            df = pd.read_csv(self.path, **kw_options)
        elif self.type == "euro_csv":
            df = pd.read_csv(self.path, sep=";", decimal=",", **kw_options)
        else:
            raise TypeError(
                'Can only read csv files as dataframes. Use method "get_raw" for raw data'
            )

        return df

    def get_raw(self, params: Dict = None, buffer: bool = False) -> Raw:
        """Get data as raw (unstructured) data.

        Example::

            data_sources["my_raw_datasource"].get_raw()

        :param params: Currently not implemented
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :return: The file's bytes (binary) or string (text) contents, possibly cached according to config value of `expires:`
        :rtype: bytes or str
        """
        if buffer:
            raise NotImplementedError("Buffered reading not supported yet")

        kw_options = self.options

        logger.debug(
            "Loading raw {} {} with options {}...".format(
                self.type, self.path, kw_options
            )
        )

        raw: Raw
        if self.type == "text_file":
            with open(self.path, "r") as txt_file:
                raw = txt_file.read(**kw_options)
        elif self.type == "binary_file":
            with open(self.path, "rb") as bin_file:
                raw = bin_file.read(**kw_options)
        else:
            raise TypeError(
                "Can only read binary data or text strings as raw file. "
                'Use method "get_dataframe" for dataframes'
            )

        return raw


class DataSink:
    """Interface, used by the Data Scientist's model to persist data (usually prediction results).
    Concrete DataSinks (for files, data bases, etc.) need to inherit from this class.
    """

    serves: List[str] = []

    def __init__(
        self,
        identifier: str,
        datasink_config: Dict,
        sub_config: Optional[Dict] = None,
    ):
        """Please call super().__init(...) when overwriting this method
        """
        self.id = identifier
        self.config = datasink_config
        self.options = self.config.get("options", {})

    @abc.abstractmethod
    def put_dataframe(
        self, dataframe: pd.DataFrame, params=None, buffer=False
    ) -> None:
        ...

    @abc.abstractmethod
    def put_raw(self, raw_data: Raw, params=None, buffer=False) -> None:
        ...

    def __del__(self):
        """Overwrite to clean up any resources (connections, temp files, etc.).
        """
        ...


class FileDataSink(DataSink):
    """DataSink for putting data into files.

    See :attr:`serves` for the available types.

    Configuration example::

        datasinks:
          # ... (other datasinks)
          my_datasink:
            type: euro_csv  # `euro_csv` changes separators to ";" and decimals to "," w.r.t. `csv`
            path: /some/file.csv  # Can be URL, uses `df.to_csv` internally
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when fetching the data using `df.to_csv`
          my_raw_datasink:
            type: text_file  # raw files can also be of type `binary_file`
            path: /some/file.txt  # Can be URL
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when writing the data using `fh.write`
    """

    serves = SUPPORTED_FILE_TYPES

    def __init__(self, identifier: str, datasink_config: Dict):
        super().__init__(identifier, datasink_config)

        ds_type = datasink_config["type"]
        if ds_type not in SUPPORTED_FILE_TYPES:
            raise TypeError(
                "{} is not a datasink file type (in datasink {}).".format(
                    repr(ds_type), repr(identifier)
                )
            )

        self.type = ds_type
        self.path = datasink_config["path"]

    def put_dataframe(
        self,
        dataframe: pd.DataFrame,
        params: Dict = None,
        buffer: bool = False,
    ) -> None:
        """Write a pandas dataframe to file.
        The default is not to save the dataframe's row index.
        Configure the DataSink's `options` dict to pass keyword arguments to `my_df.to_csv`.

        Example::

            data_sinks["my_datasink"].put_dataframe(my_df)

        :param dataframe: The pandas dataframe to save
        :type dataframe: pandas DataFrame
        :param params: Currently not implemented
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool
        """
        if buffer:
            raise NotImplementedError("Buffered writing not supported yet")

        kw_options = self.options
        if "index" not in kw_options:
            kw_options["index"] = False

        logger.debug(
            "Writing dataframe to type {} file {} with options {}...".format(
                self.type, self.path, kw_options
            )
        )
        if self.type == "csv":
            dataframe.to_csv(self.path, **kw_options)
        elif self.type == "euro_csv":
            dataframe.to_csv(self.path, sep=";", decimal=",", **kw_options)
        else:
            raise TypeError(
                'Can only write dataframes to csv file. Use method "put_raw" for raw data'
            )

    def put_raw(
        self, raw_data: Raw, params: Dict = None, buffer: bool = False,
    ) -> None:
        """Write raw (unstructured) data to file.

        Example::

            data_sinks["my_raw_datasink"].put_raw(my_data)

        :param raw_data: The data to save (bytes for binary, string for text file)
        :type raw_data: bytes or str
        :param params: Currently not implemented
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool
        """
        if buffer:
            raise NotImplementedError("Buffered writing not supported yet")

        kw_options = self.options

        logger.debug(
            "Writing raw {} file {} with options {}...".format(
                self.type, self.path, kw_options
            )
        )
        if self.type == "text_file":
            with open(self.path, "w", **kw_options) as txt_file:
                raw_str: str = cast(str, raw_data)
                txt_file.write(raw_str)
        elif self.type == "binary_file":
            with open(self.path, "wb", **kw_options) as bin_file:
                raw_bytes: bytes = cast(bytes, raw_data)
                bin_file.write(raw_bytes)
        else:
            raise TypeError(
                "Can only write binary data or text strings as raw file. "
                + 'Use method "put_dataframe" for dataframes'
            )


class OracleDataSink(DataSink):
    """DataSink for Oracle database connections.

    Creates a long-living connection on initialization.

    Configuration example::

        dbms:
          # ... (other connections)
          my_connection:  # NOTE: You can use the same connection for several datasources and datasinks
            type: oracle
            host: host.example.com
            port: 1251
            user_var: MY_USER_ENV_VAR
            password_var: MY_PW_ENV_VAR  # optional
            service_name: servicename.example.com
            options: {}  # used as **kwargs when initializing the DB connection
        # ...
        datasinks:
          # ... (other datasinks)
          my_datasink:
            type: dbms.my_connection
            table: somewhere.my_table
            tags: [train] # generic parameter, see documentation on DataSources and DataSinks
            options: {}   # used as **kwargs when fetching the query using `pandas.to_sql`
    """

    serves = ["dbms.oracle"]

    def __init__(
        self, identifier: str, datasink_config: Dict, dbms_config: Dict
    ):
        super().__init__(identifier, datasink_config)

        self.dbms_config = dbms_config

        logger.info(
            "Establishing Oracle database connection for datasource {}...".format(
                self.id
            )
        )

        self.connection = _get_oracle_connection(dbms_config)

    def put_dataframe(
        self,
        dataframe: pd.DataFrame,
        params: Dict = None,
        buffer: bool = False,
    ) -> None:
        """Store the pandas dataframe as a table.
        The default is not to store the dataframe's row index.
        Configure the DataSink's options dict to pass keyword arguments to `df.o_sql`.

        Example::

            data_sinks["my_datasink"].put_dataframe(my_df)

        :param dataframe: The pandas dataframe to store
        :type dataframe: pandas DataFrame
        :param params: Currently not implemented
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool
        """
        if buffer:
            raise NotImplementedError("Buffered storing not supported yet")

        # TODO: maybe want to open/close connection on every method call (shouldn't happen often)
        table = self.config["table"]
        kw_options = self.options
        if "index" not in kw_options:
            kw_options["index"] = False

        logger.debug(
            "Storing data in table {} with options {}...".format(
                table, kw_options
            )
        )
        dataframe.to_sql(table, con=self.connection, **kw_options)

    def put_raw(
        self, raw_data, params: str = None, buffer: bool = False
    ) -> None:
        """Not implemented.

        :raises NotImplementedError: Raw/blob format currently not supported.
        """
        raise NotImplementedError(
            "OracleDataSink currently does not not support raw format/blobs. "
            'Use method "put_dataframe" for raw data'
        )

    def __del__(self):
        if hasattr(self, "connection"):
            self.connection.close()


def to_plain_python_obj(possible_ndarray):
    if isinstance(possible_ndarray, dict):
        return {
            key: to_plain_python_obj(val)
            for key, val in possible_ndarray.items()
        }
    if isinstance(possible_ndarray, np.int64):
        return int(possible_ndarray)
    if isinstance(possible_ndarray, np.float32):
        return float(possible_ndarray)
    elif isinstance(possible_ndarray, list) or isinstance(
        possible_ndarray, tuple
    ):
        return [to_plain_python_obj(val) for val in possible_ndarray]
    elif isinstance(possible_ndarray, np.ndarray):
        logger.debug("Automatically converting ndarray to plain python list")
        return possible_ndarray.tolist()
    elif isinstance(possible_ndarray, pd.DataFrame):
        logger.debug("Automatically converting DataFrame to plain python dict")
        return possible_ndarray.to_dict()
    else:
        return possible_ndarray


_order_columns_called = 0


def order_columns(obj: Union[pd.DataFrame, np.ndarray, Dict]):
    """Order the columns of a DataFrame, a dict, or a Numpy structured array.
    Use this on your training data right before passing it into the model.
    This will guarantee that the model is trained with a reproducible column order.

    Same in your test code.

    Most importantly, use this also in your `predict` method, as the incoming
    `args_dict` does not have a deterministic order.

    Params:
        obj: a DataFrame, a dict, or a Numpy structured array

    Returns:
        The `obj` with columns ordered lexicographically
    """
    # Implementation note: not using singledispatch due to necessary checks for all calls regardless of type.
    # try:
    #     caller = sys._getframe().f_back.f_code.co_name
    # except AttributeError:
    #     caller = None
    global _order_columns_called
    _order_columns_called += 1

    if isinstance(obj, pd.DataFrame):
        cols_sorted = sorted(obj.columns.tolist())
        return obj.loc[:, cols_sorted]
    elif isinstance(obj, np.ndarray):
        if obj.dtype.names:
            cols_sorted = sorted(obj.dtype.names)
            return obj[cols_sorted]
        else:
            raise TypeError(
                "Non-structured numpy array does not have column names that can be ordered."
            )
    elif isinstance(obj, dict):
        ordered = OrderedDict(sorted(obj.items()))
        return ordered
    else:
        raise TypeError(
            "`order_columns` called on unsupported type: {}".format(type(obj))
        )



import logging
from typing import Dict, Optional

from mllaunchpad.resource import DataSource, get_user_pw


logger = logging.getLogger(__name__)

try:
    import records
except ModuleNotFoundError:
    logger.warning("Please install the Records package to be able to use RecordsDbDataSource.")


class RecordsDbDataSource(DataSource):
    """DataSource for a bunch of relational database types:
    RedShift, Postgres, MySQL, SQLite, Oracle, Microsoft SQL.

    EXPERIMENTAL

    See :attr:`serves` for the available types.

    Creates a long-living connection on initialization.

    Configuration::

        dbms:
          # ... (other connections)
          my_connection:  # NOTE: You can use the same connection for several datasources and datasinks
            type: oracle  # see attribute serves for available types
            host: host.example.com
            port: 1251  # optional
            user_var: MY_USER_ENV_VAR
            password_var: MY_PW_ENV_VAR  # optional
            service_name: servicename.example.com  # optional
            options: {}  # used as **kwargs when initializing the DB connection
        # ...
        datasources:
          # ... (other datasources)
          my_datasource:
            type: dbms.my_connection
            query: SELECT * FROM users.my_table where id = :id  # fill `:params` by calling `get_dataframe` with a `dict`
            expires: 0     # generic parameter, see documentation on DataSources
            tags: [train]  # generic parameter, see documentation on DataSources and DataSinks
    """

    serves = [
        "dbms.oracle",
        "dbms.redshift",
        "dbms.postgres",
        "dbms.mysql",
        "dbms.sqlite",
        "dbms.ms_sql",
    ]

    def __init__(self, identifier: str, datasource_config: Dict, dbms_config: Optional[Dict]):
        super().__init__(identifier, datasource_config)

        self.dbms_config = dbms_config

        logger.info(
            "Establishing Records database connection for datasource {}...".format(
                self.id
            )
        )
        # if "connect" not in dbms_config:
        #     raise ValueError(f'No connection string (property "connect") in datasource {self.id} config')
        dbtype = dbms_config["type"]
        user, pw = get_user_pw(dbms_config["user_var"], dbms_config["password_var"])
        host = dbms_config["host"]
        port = ":" + str(dbms_config["port"]) if "port" in dbms_config else ""
        service_name = (
            "/?service_name=" + dbms_config["service_name"]
            if "service_name" in dbms_config
            else ""
        )
        kwargs = dbms_config.get("options", {})
        connection_string = f"{dbtype}://{user}:{pw}@{host}{port}{service_name}"
        self.db = records.Database(connection_string, **kwargs)

    def get_dataframe(self, params=None, buffer=False):
        """Get data as a pandas dataframe.

        Example::

            data_sources["my_datasource"].get_dataframe({"id": 387})

        :param params: Query parameters to fill in query (e.g. `:id` with value 387)
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :return: DataFrame object, possibly cached according to expires-config
        """
        if buffer:
            raise NotImplementedError("Buffered reading not supported yet")
            # the resulting `rows` of a query provides a nice way to do this, though

        query = self.config["query"]
        params = params or {}

        logger.debug(
            "Fetching query {} with params {}...".format(
                query, params
            )
        )
        rows = self.db.query(query, fetchall=True, **params)
        df = rows.export("df")

        return df

    def get_raw(self, params=None, buffer=False):
        """Not implemented.

        :param params: Query parameters to fill in query (e.g. `:id` with value 387)
        :type params: optional dict
        :param buffer: Currently not implemented
        :type buffer: optional bool

        :raises NotImplementedError:

        :return: Nothing, always raises NotImplementedError
        """
        raise NotImplementedError("Raw data not supported")

    # def __del__(self):
    #     if hasattr(self, "db"):
    #         self.db.close()



# Stdlib imports
import abc
import logging


logger = logging.getLogger(__name__)


class ModelMakerInterface(abc.ABC):
    """Abstract model factory interface for Data-Scientist-created models.
    Please inherit from this class and put your training code into the
    method "create_trained_model". This method will be called by the framework when
    your model needs to be (re)trained.

    Why not simply use static methods?
        We want to make it possible for create_trained_model to pass extra info
        test_trained_model without extending the latter with optional keyword
        arguments that might be confusing for the 90% of cases where they are
        not needed. So we rely on the smarts of the person inheriting from this
        class to find a solution/shortcuts if they want to do more difficult
        things e.g. want to do the train/test split themselves.
    """

    @abc.abstractmethod
    def create_trained_model(
        self, model_conf, data_sources, data_sinks, old_model=None
    ):
        """Implement this method, including data prep/feature creation.
        No need to test your model here. Put testing code in test_trained_model, which
        will be called automatically after training.
        (Feel free to put common code for preparing data into another function,
        class, library, ...)

        Params:
            model_conf:   the model configuration dict from the config file
            data_sources: dict containing the data sources (this includes your train/validation/test data), as configured in the config file
            data_sinks:   dict containing the data sinks, as configured in the config file. Usually unused when training.
            old_model:    contains an old model, if it exists, which can be used for incremental training. default: None

        Return:
            The trained model/data/anything which you want to use in the predict()
            function. (usually simply a fitted model object, but can be anything,
            like a dict of several models, a model with some extra info, etc.)
            (Whatever you return here gets automatically stuffed into your
            ModelInterface-inherited object and is accessible there using
            predict's model parameter (or the self.contents attribute.))
        """

    @abc.abstractmethod
    def test_trained_model(self, model_conf, data_sources, data_sinks, model):
        """Implement this method, including data prep/feature creation.
        This method will be called to re-test a model, e.g. to check whether
        it has to be re-trained.
        (Feel free to put common code for preparing data into another function,
        class, library, ...)

        Params:
            model_conf:   the model configuration dict from the config file
            data_sources: dict containing the data sources (this includes your train/validation/test data), as configured in the config file
            data_sinks:   dict containing the data sinks, as configured in the config file. Usually unused when testing.
            model:        your model object (whatever you returned in create_trained_model)

        Return:
            Return a dict of metrics (like 'accuracy', 'f1', 'confusion_matrix', etc.)
        """


class ModelInterface(abc.ABC):
    """Abstract model interface for Data-Scientist-created models.
    Please inherit from this class when creating your model to make
    it usable for ModelApi.

    You don't need to create this object yourself when training.
    It is created automatically and the model/info returned from create_trained_model
    is made accessible to you through the self.contents attribute.
    """

    def __init__(self, contents=None):
        """If you overwrite __init__, please call super().__init__(...)
        at the beginning. Otherwise, you need to assign self.contents to the
        contents parameter manually in __init__.

        Params:
            contents:  any object that is needed for prediction (usually a trained
                      classifier or predictor). It is passed to the predict()
                      method as the "model" parameter for convenience.
        """
        self.contents = contents

    @abc.abstractmethod
    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):
        """Implement this method, including data prep/feature creation based on argsDict.
        argsDict can also contain an id which the model can use to fetch data
        from any data_sources.
        (Feel free to put common code for preparing data into another function,
        class, library, ...)

        Params:
            model_conf:   the model configuration dict from the config file
            data_sources: dict containing the data sources
            data_sinks:   dict containing the data sinks, as configured in the config file.
            model:        your model object (whatever you returned in create_trained_model)
            argsDict:     parameters the API was called with, dict of strings (any type conversion needs to be done by you)

        Return:
            Prediction result as a dictionary/list structure which will be automatically turned into JSON.
        """

    def __del__(self):
        """Clean up any resources (temporary files, sockets, etc.).
        If you overwrite this method, please call super().__del__() at the beginning.
        """



"""This module contains the WSGI app to start using a WSGI server like
   e.g. gunicorn or Apache with mod_wsgi.

Example:
    `$ gunicorn -w 4 --bind 127.0.0.1:5000 mllaunchpad.wsgi:application`
"""

# Stdlib imports
import logging

# Third-party imports
from flask import Flask

# Project imports
from mllaunchpad import config, logutil
from mllaunchpad.api import ModelApi


logutil.init_logging()
logger = logging.getLogger(__name__)

# In order to be able to generate API docs automatically, it is unfortunately
# necessary to wrap the preparatory code in a try:except: statement.
try:
    conf = config.get_validated_config()
except FileNotFoundError:
    logger.error(
        "Config file could not be loaded. Starting the Flask application "
        "will fail."
    )
    conf = None

if conf:
    # if you change the name of the application variable, you need to
    # specify it explicitly for gunicorn: gunicorn ... launchpad.wsgi:appname
    application = Flask(__name__, root_path=conf["api"].get("root_path"))
    ModelApi(conf, application)

    if __name__ == "__main__":  # pragma: no cover
        logger.warning(
            "Starting Flask debug server.\nIn production, please use a WSGI server, "
            "e.g. 'gunicorn -w 4 -b 127.0.0.1:5000 mllaunchpad.wsgi:application'"
        )
        # Flask apps must not be run in debug mode in production, because this allows for arbitrary code execution.
        # We know that and advise the user that this is only for debugging, so this is not a security issue (marked nosec):
        application.run(debug=True)  # nosec

# To start an instance of production server with 4 workers:
#  1. Set environment variables if required
#  2. gunicorn --workers 4 --bind 127.0.0.1:5000 launchpad.wsgi
#
# Gunicorn is a WSGI server - usable for prod

# For performance/load balancing, use it with a reverse HTTP proxy like nginx:
# https://gunicorn.org/#deployment
# server {
#     listen 80;
#     server_name example.org;
#     access_log  /var/log/nginx/example.log;
#
#     location / {
#         proxy_pass http://127.0.0.1:8000;
#         proxy_set_header Host $host;
#         proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
#     }
#   }



"""Top-level package for ML Launchpad."""

# Third-party imports
import pkg_resources

# Project imports
from mllaunchpad.config import get_validated_config
from mllaunchpad.model_actions import predict, retest, train_model
from mllaunchpad.model_interface import ModelInterface, ModelMakerInterface
from mllaunchpad.resource import order_columns


__version__ = pkg_resources.get_distribution("mllaunchpad").version


# Get rid of stupid pyflakes warning (be warned: redundancy warning):
__all__ = [
    "train_model",
    "retest",
    "predict",
    "get_validated_config",
    "ModelInterface",
    "ModelMakerInterface",
    "order_columns",
]



# Stdlib imports
import os

# Third-party imports
import yaml


class Loader(yaml.SafeLoader):
    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]

        super().__init__(stream)

    def include(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))

        with open(filename, "r") as f:
            # Normally, one should use safe_load(), but our Loader
            # is a subclass of yaml.SafeLoader
            return yaml.load(f, Loader)  # nosec


Loader.add_constructor("!include", Loader.include)



"""Convenience functions for executing training, testing and prediction"""

# Stdlib imports
import logging
import sys
from typing import Dict

# Project imports
from mllaunchpad import resource
from mllaunchpad.model_interface import ModelInterface, ModelMakerInterface


logger = logging.getLogger(__name__)


_cached_model_stores: Dict = {}
_cached_model_tuples: Dict = {}
_cached_data_source_sink_tuples: Dict = {}
_cached_model_makers: Dict = {}
_cached_model_classes: Dict = {}

suppress_order_columns_not_used_warning = (
    "To adjust this warning, set model config's `order_columns_not_used_warning` "
    "to `always` (default), `test_and_predict` or `never`."
)


def train_model(complete_conf, cache=True):
    """Train and test a model as specified in the configuration and persist
    it in the model store

    Params:
        complete_conf: configuration dict
        cache:         (default: True) whether to cache the data sources/sinks and helper objects (cache lookup is done by model name and model version). If in doubt, leave at default.

    Returns:
        (model, metrics) tuple
    """
    logger.debug("Creating trained model...")
    user_mm = _get_model_maker(complete_conf, cache=cache)

    dso, dsi = _get_data_sources_and_sinks(
        complete_conf, tags=["train", "test"], cache=cache
    )

    model_conf = complete_conf["model"]

    try:
        logger.debug("Trying to load old model...")
        old_model_wrapper, _ = _get_model(complete_conf, cache=cache)
        old_inner_model = old_model_wrapper.contents
        # logger.info('Loaded old model %s %s', meta['name'], meta['version'])
    except FileNotFoundError:
        logger.info("No old model to load")
        old_inner_model = None

    inner_model = user_mm.create_trained_model(
        model_conf, dso, dsi, old_model=old_inner_model
    )
    m_cls = _get_model_class(complete_conf, cache=cache)
    model_wrapper: ModelInterface = m_cls(contents=inner_model)

    if resource._order_columns_called:
        model_wrapper.__ordered_columns = True
    elif (
        "order_columns_not_used_warning" not in model_conf
        or model_conf["order_columns_not_used_warning"] == "always"
    ):
        logger.warning(
            "Training code does not call function order_columns. "
            + suppress_order_columns_not_used_warning
        )

    if not isinstance(model_wrapper, ModelInterface):
        logger.warning(
            "Model's class is not a subclass of ModelInterface: %s",
            model_wrapper,
        )

    logger.debug("Testing trained model...")
    metrics = user_mm.test_trained_model(model_conf, dso, dsi, inner_model)
    _check_ordered_columns(
        complete_conf, model_wrapper, "testing code", times=2
    )

    model_store = _get_model_store(complete_conf, cache=cache)
    model_store.dump_trained_model(complete_conf, model_wrapper, metrics)

    logger.info(
        "Created and stored trained model %s, version %s, metrics %s",
        model_conf["name"],
        model_conf["version"],
        metrics,
    )

    return model_wrapper, metrics


def retest(complete_conf, cache=True):
    """Retest a model as specified in the configuration and persist its
    test metrics in the model store

    Params:
        complete_conf: configuration dict
        cache:         (default: True) whether to cache the data sources/sinks and helper objects (cache lookup is done by model name and model version). If in doubt, leave at default.

    Returns:
        test_metrics
    """
    logger.debug("Retesting existing trained model...")
    user_mm = _get_model_maker(complete_conf, cache=cache)

    dso, dsi = _get_data_sources_and_sinks(
        complete_conf, tags="test", cache=cache
    )

    model_wrapper, _ = _get_model(complete_conf, cache=cache)
    inner_model = model_wrapper.contents

    model_conf = complete_conf["model"]
    test_metrics = user_mm.test_trained_model(
        model_conf, dso, dsi, inner_model
    )
    _check_ordered_columns(complete_conf, model_wrapper, "retesting code")

    model_store = _get_model_store(complete_conf, cache=cache)
    model_store.update_model_metrics(model_conf, test_metrics)

    logger.info(
        "Retested existing model %s, version %s, new metrics %s",
        model_conf["name"],
        model_conf["version"],
        test_metrics,
    )

    return test_metrics


def predict(complete_conf, arg_dict=None, cache=True):
    """Carry out prediction for the model specified in the configuration

    Params:
        complete_conf: configuration dict
        arg_dict:      arguments dict for the prediction (analogous to what it would get from a web API)
        cache:         (default: True) whether to cache the data sources/sinks and helper objects (cache lookup is done by model name and model version). If in doubt, leave at default.

    Returns:
        model's prediction output
    """
    logger.debug("Applying model for prediction...")

    dso, dsi = _get_data_sources_and_sinks(
        complete_conf, tags="predict", cache=cache
    )

    model_wrapper, _ = _get_model(complete_conf, cache=cache)
    inner_model = model_wrapper.contents

    model_conf = complete_conf["model"]
    output = model_wrapper.predict(
        model_conf, dso, dsi, inner_model, arg_dict or {}
    )
    _check_ordered_columns(complete_conf, model_wrapper, "prediction code")

    output = resource.to_plain_python_obj(output)

    return output


def clear_caches():
    global _cached_model_stores
    global _cached_model_tuples
    global _cached_data_source_sink_tuples
    global _cached_model_makers
    global _cached_model_classes

    _cached_model_stores.clear()
    _cached_model_tuples.clear()
    _cached_data_source_sink_tuples.clear()
    _cached_model_makers.clear()
    _cached_model_classes.clear()


def _model_key(model_conf):
    return model_conf["name"] + "_" + model_conf["version"]


def _find_subclass(module_name, superclass):
    """Locate and instantiate class of data-scientist-provided ModelMaker
    or model class.
    For this to work, your model module (.py file) needs to be in python's
    sys.path (usually the case).
    Also set the config's model: module property accordingly.

    Params:
        module_name: the name of the module
        superclass: subclasses of which class to locate

    Returns:
        subclass as found in data scientist's module
    """

    if "." not in sys.path:
        sys.path.append(".")

    __import__(module_name)

    classes = superclass.__subclasses__()
    print(classes)
    if len(classes) != 1:
        raise ValueError(
            "The configured model module (.py file) must contain "
            + "one {}-inheriting class definition, but contains {}.".format(
                superclass, len(classes)
            )
        )

    cls = classes[0]
    logger.debug("Found %s class named %s", superclass, cls)

    return cls


def _get_model_maker(complete_conf, cache=True):
    """Locate and instantiate class of data-scientist-provided ModelMaker
    (which the data scientist inherited from ModelMakerInterface).
    For this to work, your model module (.py file) needs to be in python's
    sys.path (usually the case).
    Also set the config's model: module property accordingly.

    Params:
        complete_conf: the configuration dict

    Returns:
        instance of data scientist's ModelMaker model factory object
    """
    global _cached_model_makers

    key = _model_key(complete_conf["model"])

    mm = _cached_model_makers.get(key)
    if mm is None:
        logger.debug("Locating and instantiating ModelMaker...")
        mm_cls = _find_subclass(
            complete_conf["model"]["module"], ModelMakerInterface
        )
        mm = mm_cls()
        if cache:
            _cached_model_makers[key] = mm
        logger.debug("Instantiated ModelMaker object %s", mm)

    return mm


def _get_model_class(complete_conf, cache=True):
    """Locate class of data-scientist-provided model
    (which the data scientist inherited from ModelInterface).
    For this to work, your model module (.py file) needs to be in python's
    sys.path (usually the case).
    Also set the config's model: module property accordingly.

    Params:
        complete_conf: the configuration dict

    Returns:
        data scientist's Model class
    """
    global _cached_model_classes

    key = _model_key(complete_conf["model"])

    m_cls = _cached_model_classes.get(key)
    if m_cls is None:
        logger.debug("Locating Model class...")
        m_cls = _find_subclass(
            complete_conf["model"]["module"], ModelInterface
        )
        if cache:
            _cached_model_classes[key] = m_cls

    return m_cls


def _get_model_store(complete_conf, cache=True):
    global _cached_model_stores

    key = _model_key(complete_conf["model"])

    model_store = _cached_model_stores.get(key) or resource.ModelStore(
        complete_conf
    )

    if cache:
        _cached_model_stores[key] = model_store

    return model_store


def _get_model(complete_conf, cache=True):
    global _cached_model_tuples
    model_store = _get_model_store(complete_conf, cache=cache)

    model_conf = complete_conf["model"]
    key = _model_key(model_conf)

    model_tuple = _cached_model_tuples.get(key)
    if model_tuple is None:
        logger.info("Loading model...")
        model_wrapper, meta = model_store.load_trained_model(model_conf)
        logger.info(
            "Model loaded: {}, version: {}, created {}".format(
                meta["name"], meta["version"], meta["created"]
            )
        )

        model_tuple = (model_wrapper, meta)

        if cache:
            _cached_model_tuples[key] = model_tuple

    return model_tuple


def _get_data_sources_and_sinks(complete_conf, tags=None, cache=True):
    global _cached_data_source_sink_tuples

    key = _model_key(complete_conf["model"])

    ds_tuple = _cached_data_source_sink_tuples.get(key)
    if ds_tuple is None:
        logger.info("Initializing datasources...")
        dso, dsi = resource.create_data_sources_and_sinks(
            complete_conf, tags=tags
        )
        logger.info(
            "%s datasource(s) initialized: %s", len(dso), list(dso.keys())
        )
        logger.info(
            "%s datasink(s) initialized: %s", len(dsi), list(dsi.keys())
        )

        ds_tuple = (dso, dsi)

        if cache:
            _cached_data_source_sink_tuples[key] = ds_tuple

    return ds_tuple


def _check_ordered_columns(complete_conf, model_wrapper, what: str, times=1):
    if (
        "order_columns_not_used_warning" in complete_conf["model"]
        and complete_conf["model"]["order_columns_not_used_warning"] == "never"
    ):
        return
    if hasattr(model_wrapper, "__ordered_columns"):
        if resource._order_columns_called < times:
            logger.warning(
                "Model has been trained on ordered columns, but "
                "{} does not call function order_columns. {}".format(
                    what, suppress_order_columns_not_used_warning
                )
            )



import logging

from mllaunchpad import ModelInterface, ModelMakerInterface
import pandas as pd
from sklearn import tree
from sklearn.metrics import accuracy_score, confusion_matrix


logger = logging.getLogger(__name__)

# Train this example from the command line:
# python -m mllaunchpad -c complex_cfg.yml train
#
# Start REST API:
# python -m mllaunchpad -c complex_cfg.yml api
#
# Example API call:
# http://127.0.0.1:5000/guessiris/v0/somethings?x=3&sepal.length=4.9&sepal.width=2.4&petal.length=3.3&petal.width=1


def data_prep(X):
    # prepping features, maybe parsing strange formats, imputing, cleaning text ...
    return X


class MyModelMaker(ModelMakerInterface):
    """
    """

    def create_trained_model(self, model_conf, data_sources, data_sinks, old_model=None):
        # demo: get the database data source
        limit = model_conf["train_options"]["num_ora_rows"]
        dbdf = data_sources["panel"].get_dataframe(params={"limit": limit})
        print(dbdf)

        # just for lolz
        number_to_add = model_conf["train_options"]["magic_number"]
        my_lame_predictor = lambda x: x + number_to_add

        # train a tree as a demo
        df = data_sources["petals"].get_dataframe()
        X_train = df.drop("variety", axis=1)
        y_train = df["variety"]

        # optional data prep/feature creation/refinement here...
        my_tree = tree.DecisionTreeClassifier()
        my_tree.fit(X_train, y_train)

        model = {"lame_pred": my_lame_predictor, "petal_pred": my_tree}

        return model

    def test_trained_model(self, model_conf, data_sources, data_sinks, model):
        df = data_sources["petals_test"].get_dataframe()
        X_test = df.drop("variety", axis=1)
        y_test = df["variety"]

        my_tree = model["petal_pred"]

        y_predict = my_tree.predict(X_test)

        acc = accuracy_score(y_test, y_predict)
        conf = confusion_matrix(y_test, y_predict).tolist()
        metrics = {"accuracy": acc, "confusion_matrix": conf}

        return metrics


class MyModel(ModelInterface):
    """Does some simple prediction
    """

    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):
        logger.info("Hey, look at me -- I'm carrying out a prediction")

        # Do some lame prediction (= addition)
        x_raw = args_dict["x"]

        # optional data prep/feature creation for x here...
        x = data_prep(x_raw)

        name_df = data_sources["first_names"].get_dataframe()
        random_name = name_df.sample(n=1)["name"].values[0]

        lame_predictor = ModelInterface()["lame_pred"]
        y = lame_predictor(x)

        # Also try iris petal-based prediction:
        petal_predictor = model["petal_pred"]
        X2 = pd.DataFrame(
            {
                "sepal.length": [args_dict["sepal.length"]],
                "sepal.width": [args_dict["sepal.width"]],
                "petal.length": [args_dict["petal.length"]],
                "petal.width": [args_dict["petal.width"]],
            }
        )
        y2 = petal_predictor.predict(X2)[0]

        return {"the_result_yo": y, "random_name": random_name, "iris_variety": y2}



"""This module contains functionality for reading and validating the
   configuration.
"""

# Stdlib imports
import logging
import os
from typing import Dict
from warnings import warn

# Third-party imports
import yaml  # https://camel.readthedocs.io/en/latest/yamlref.html

# Project imports
import mllaunchpad
from mllaunchpad.yaml_loader import Loader


logger = logging.getLogger(__name__)

CONFIG_DEFAULT = "./LAUNCHPAD_CFG.yml"
CONFIG_ENV = os.environ.get("LAUNCHPAD_CFG", CONFIG_DEFAULT)
required_config: Dict[str, Dict] = {
    # datasources and datasinks are optional
    "model_store": {"location": {}},
    "model": {
        "name": {},
        "version": {},
        "module": {},
        # custom model keys are optional
    },
    # api: is optional because using mllaunchpad's WSGI/Flask API is optional
    # (e.g. when using mllaunchpad's convenience functions in creating Azure functions).
    # You are free to add more custom settings in the config file. For an example, see
    # https://github.com/schuderer/mllaunchpad-template
}


def validate_config(config_dict, required, path=""):
    for item in required:
        path_start = (path + ":") if path else ""
        if item not in config_dict:
            raise ValueError(
                "Missing key in config file: {}".format(path_start + item)
            )
        validate_config(config_dict[item], required[item], path_start + item)


def check_semantics(config_dict):
    if "api" in config_dict and "version" in config_dict["api"]:
        if mllaunchpad.__version__ < "1.0.0":
            warn(
                "Specifying 'version' in the config's 'api' section is "
                "deprecated and will lead to an error in mllaunchpad>=1.0.0. "
                "Specify 'version' in the 'model' section instead. "
                "Your 'api:version' value will be ignored.",
                DeprecationWarning,
            )
            del config_dict["api"]["version"]
        else:
            raise ValueError(
                "'api:version:' is not allowed in the config, "
                "only 'model:version:'."
            )


def get_validated_config(filename=CONFIG_ENV):
    """Read the configuration from file and return it as a dict object.

    Validation is not done yet and planned when the configuration is more
    stable.

    Params:
        - filename: path to configuration file
                    (default: environment variable LAUNCHPAD_CFG or
                     file LAUNCHPAD_CFG.yml)

    Returns:
        dict with configuration
    """
    if filename == CONFIG_DEFAULT:
        logger.warning(
            "Config filename environment variable LAUNCHPAD_CFG not set, "
            "using default file: %s",
            repr(CONFIG_DEFAULT),
        )
    logger.info("Loading configuration file %s...", filename)

    with open(filename) as f:
        # Normally, one should use safe_load(), but our Loader
        # is a subclass of yaml.SafeLoader
        y = yaml.load(f, Loader)  # nosec

    validate_config(y, required_config)
    check_semantics(y)

    logger.debug("Configuration loaded and validated: %s", y)

    return y



"""This module provides the command line interface for ML Launchpad"""

# Stdlib imports
import json
import sys
from logging import Logger
from typing import Dict

# Third-party imports
import click
from flask import Flask

# Project imports
import mllaunchpad as mllp
from mllaunchpad import logutil
from mllaunchpad.api import ModelApi, generate_raml


# Fix for click using the wrong name if run using `python -m mllaunchpad`
# Also see: https://github.com/pallets/click/issues/365
sys.argv[0] = mllp.__name__


# Adapted from https://click.palletsprojects.com/en/7.x/advanced/
class AliasedGroup(click.Group):
    """Commands can be abbreviated, e.g. t or tr for train, a for api, etc."""

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [
            x for x in self.list_commands(ctx) if x.startswith(cmd_name)
        ]
        if not matches:
            return None
        elif len(matches) == 1:
            click.echo("Command {} matches {}".format(cmd_name, matches[0]))
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(
            "Command {} is ambiguous: {}".format(
                cmd_name, ", ".join(sorted(matches))
            )
        )


# TODO: Migrate to @dataclass when dropping support for Python 3.6
class Settings:
    def __init__(self):
        self.verbose: bool = False
        self.logger: Logger = None
        self.conf_file: str = None
        self._config: Dict = None

    @property
    def config(self):
        if not hasattr(self, "_config") or not self._config:
            if self.conf_file:
                self._config = mllp.get_validated_config(self.conf_file)
            else:
                self._config = mllp.get_validated_config()
        return self._config


pass_settings = click.make_pass_decorator(Settings, ensure=True)


@click.group(
    cls=AliasedGroup, context_settings=dict(help_option_names=["-h", "--help"])
)
@click.version_option(prog_name="ML Launchpad")
@click.option("--verbose", "-v", is_flag=True, help="Print debug messages.")
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="Use this configuration file. [default: look for env var LAUNCHPAD_CFG or ./LAUNCHPAD_CFG.yml]",
)
@click.option(
    "--log-config",
    "-l",
    type=click.Path(exists=True),
    help="Use this log configuration file. [default: look for env var LAUNCHPAD_LOG or ./LAUNCHPAD_LOG.yml]",
)
@pass_settings
def main(settings, log_config, config, verbose):
    """Train, test or run a config file's model."""
    # Initialize logging before any library code so that mllp can log stuff
    if log_config:
        settings.logger = logutil.init_logging(log_config, verbose=verbose)
    else:
        settings.logger = logutil.init_logging(verbose=verbose)
    settings.verbose = verbose
    settings.conf_file = config


@main.command()
@pass_settings
def train(settings):
    """Run training, store created model and metrics."""
    _, metrics = mllp.train_model(settings.config)
    print(metrics)


@main.command()
@pass_settings
def retest(settings):
    """Retest existing model, update metrics."""
    metrics = mllp.retest(settings.config)
    print(metrics)


@main.command()
@pass_settings
def api(settings):
    """Run API server in unsafe debug mode."""
    settings.logger.warning(
        "Starting Flask debug server. In production, please "
        "use a WSGI server, e.g.\n"
        "'export LAUNCHPAD_CFG=addition_cfg.yml'\n"
        "'gunicorn -w 4 -b 127.0.0.1:5000 mllaunchpad.wsgi:application'"
    )
    app = Flask(__name__, root_path=settings.config["api"].get("root_path"))
    ModelApi(settings.config, app)
    # Flask apps must not be run in debug mode in production, because this allows for arbitrary code execution.
    # We know that and advise the user that this is only for debugging, so this is not a security issue (marked nosec):
    app.run(debug=True)  # nosec


@main.command()
@click.argument("json-file", type=click.File("r"), default=sys.stdin)
@pass_settings
def predict(settings, json_file):
    """Run prediction on features from JSON file ( - for stdin).

    \b
    Example JSON: { "petal.width": 1.4, "petal.length": 2.0,
                    "sepal.width": 1.8, "sepal.length": 4.0 }
    """
    arg_dict = json.load(json_file)
    output = mllp.predict(settings.config, arg_dict=arg_dict)
    print(output)


@main.command(name="generate-raml")
@click.argument("datasource-name", type=str, required=True)
@pass_settings
def cli_generate_raml(settings, datasource_name):
    """Generate and print RAML template from DATASOURCE_NAME.

    The datasource named DATASOURCE_NAME in the config will be used
    to create the API's query parameters (from columns), types, and examples.
    """
    print(generate_raml(settings.config, data_source_name=datasource_name))


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""This module contains functionality for generic creation
   and handling of RESTful APIs for Machine Learning Models.
   Among others, it handles parsing the RAML definition,
   and validating parameters.
"""

# Stdlib imports
import logging
import re
from collections import OrderedDict

# Third-party imports
import ramlfications
from flask_restful import Api, Resource, reqparse
from werkzeug.datastructures import FileStorage

# Project imports
from mllaunchpad import resource


logger = logging.getLogger(__name__)

# TODO: factor out datasource/model preparation code -- use model_actions instead of low-level functionality


def _get_major_api_version(config):
    match = re.match(r"\d+", config["model"]["version"])
    if match is None:
        raise ValueError(
            "Model version in configuration is malformed. Expected x.y.z, got {}".format(
                config["model"]["version"]
            )
        )
    return "v{}".format(match.group(0))


def _load_raml(config):
    api_config = config["api"]
    raml_file = api_config["raml"]
    logger.debug("Reading RAML file %s", raml_file)
    raml = ramlfications.parse(api_config["raml"])

    conf_version = _get_major_api_version(config)
    if raml.version != conf_version:
        raise ValueError(
            "API version in RAML {} does not match API version in config {}".format(
                raml.version, conf_version
            )
        )

    return raml


_type_lookup = {
    "number": float,
    "integer": int,
    "string": str,
    "boolean": bool
    # RAML Types: any, object, array, union via type expression,
    #             one of the following scalar types: number, boolean, string,
    #             date-only, time-only, datetime-only, datetime, file, integer, or nil
}


def _create_request_parser(resource_obj):
    """We only use query_params and form_params for now (no custom headers, body, etc.).
    Note that they are used interchangeably in code (so e.g. even if RAML requires
    only form_params, in reality we also accept the same params as query_params).
    """
    all_params = (resource_obj.query_params or []) + (
        resource_obj.form_params or []
    )
    added_arguments = set()
    parser = reqparse.RequestParser(bundle_errors=True)
    for p in all_params:
        if p.name in added_arguments and not p.repeat:
            raise ValueError(
                "Cannot handle RAML with multiple parameters sharing same name {}".format(
                    p.name
                )
            )

        parser.add_argument(
            p.name,
            type=_type_lookup[p.type],
            required=p.required,
            default=p.default,
            action="append" if p.repeat else "store",
            choices=p.enum,
            help=str(p.description) + " - Error: {error_msg}",
        )
        added_arguments.add(p.name)

    if (
        resource_obj.uri_params
        and resource_obj.uri_params[-1].name in added_arguments
    ):
        raise ValueError(
            'Resource URI parameter in RAML "{}" must not have same name as a parameter'.format(
                resource_obj.uri_params[-1].name
            )
        )

    if (
        resource_obj.body
        and resource_obj.body[0].mime_type == "multipart/form-data"
    ):
        # todo: how can we make sure the file mime type is correct?
        parser.add_argument("file", type=FileStorage, location="files")

    return parser


def _get_resources(raml):
    """Gets relevant resources from RAML
    """
    # only dealing with "get" method resources for now
    usable_methods = ["get", "post"]
    usable_rs = [
        r for r in raml.resources if r.method in usable_methods
    ]  # r.path == name and
    rs_without_resource_id = [
        r for r in usable_rs if not r.uri_params and not r.body
    ]
    rs_with_resource_id = [r for r in usable_rs if r.uri_params]
    rs_file_upload = [
        r
        for r in usable_rs
        if r.body and r.body[0].mime_type == "multipart/form-data"
    ]
    if (
        len(usable_rs) == 0
        or len(usable_rs) > 3
        or len(rs_without_resource_id) > 1
        or len(rs_with_resource_id) > 1
        or len(rs_file_upload) > 1
    ):
        raise ValueError(
            (
                "RAML must contain one to three resources with a method of '{}'. "
                "At most one resource each with and without uri parameter (resource id) "
                "or one file upload resource.\n"
                "There are {} resources with matching methods. Resources in RAML: {}"
            ).format(usable_methods, len(usable_rs), raml.resources)
        )

    res_normal = rs_without_resource_id[0] if rs_without_resource_id else None
    res_with_id = rs_with_resource_id[0] if rs_with_resource_id else None
    res_file = rs_file_upload[0] if rs_file_upload else None

    return res_normal, res_with_id, res_file


class QueryResource(Resource):
    # Adapted from https://flask-restful.readthedocs.io/en/latest/quickstart.html

    def __init__(self, model_api_obj, parser):
        self.model_api = model_api_obj
        self.parser = parser

    def get(self):
        args = self.parser.parse_args(
            strict=True
        )  # treats query_params and form_params as interchangeable
        logger.debug("Received GET request with arguments: %s", args)
        return self.model_api.predict_using_model(args)


class GetByIdResource(Resource):
    def __init__(self, model_api_obj, parser, id_name):
        self.model_api = model_api_obj
        self.parser = parser
        self.id_name = id_name

    def get(self, some_resource_id):
        args = self.parser.parse_args(
            strict=True
        )  # treats query_params and form_params as interchangeable
        args[self.id_name] = some_resource_id
        logger.debug(
            "Received GET request for %s %s with arguments: %s",
            self.id_name,
            some_resource_id,
            args,
        )
        return self.model_api.predict_using_model(args)


class QueryOrFileUploadResource(Resource):
    def __init__(self, model_api_obj, query_parser=None, file_parser=None):
        self.model_api = model_api_obj
        self.query_parser = query_parser
        self.file_parser = file_parser

    def get(self):
        args = self.query_parser.parse_args(
            strict=True
        )  # treat query_params and form_params as interchangeable
        logger.debug("Received GET request with arguments: %s", args)
        return self.model_api.predict_using_model(args)

    def post(self):
        if self.file_parser:
            args = self.file_parser.parse_args(strict=True)
            file_storage_obj = args["file"]
            logger.debug(
                "Received POST request with file %s of mimetype %s",
                file_storage_obj.filename,
                file_storage_obj.mimetype,
            )
        else:  # treat query_params and form_params as interchangeable
            args = self.query_parser.parse_args(strict=True)
            logger.debug("Received POST request with arguments: %s", args)
        return self.model_api.predict_using_model(args)


class ModelApi:
    """Class to plug a Data-Scientist-created model into.

    This class handles the heavy lifting of APIs for the model.

    The model is a delegate which inherits from (=implements) ModelInterface.
    It needs to provide a predict function.

    For details, see the documentation in the module `model_interface`
    """

    def __init__(self, config, application):
        """When initializing ModelApi, your model will be automatically
        retrieved from the model store based on the currently active
        configuration.

        Params:
            config:       configuration dictionary to use
            application:  flask application to use
        """
        self.model_config = config["model"]
        model_store = resource.ModelStore(config)
        self.model_wrapper = self._load_model(model_store, self.model_config)
        # Workaround (tensorflow has problem with spontaneously created threads such as with Flask):
        # https://kobkrit.com/tensor-something-is-not-an-element-of-this-graph-error-in-keras-on-flask-web-server-4173a8fe15e1
        try:
            import tensorflow as tf

            graph = tf.get_default_graph()
            self.model_wrapper.__graph = graph
        except Exception as e:
            logger.debug(
                'Optional tensorflow/flask workaround for "<tensor> is not an element of this graph" problem'
                + "resulted in: %s",
                e,
            )
        else:
            logger.info(
                'Stored tensorflow model\'s graph - tensorflow/flask workaround for "<tensor> is not an element of this graph" problem'
            )
        self.datasources, self.datasinks = self._init_datasources(config)

        logger.debug("Initializing RESTful API")
        api = Api(application)

        api_name = config["api"]["name"]
        api_version = _get_major_api_version(config)
        api_url = "/{}/{}".format(api_name, api_version)

        raml = _load_raml(config)
        res_normal, res_with_id, res_file = _get_resources(raml)

        if res_file or res_normal:
            resource_urls = {"query": None, "file": None}
            parsers = {"query": None, "file": None}
            if res_normal:
                logger.debug(
                    "Adding query resource %s to api %s",
                    res_normal.path,
                    api_url,
                )
                resource_urls["query"] = api_url + res_normal.path
                parsers["query"] = _create_request_parser(res_normal)
            if res_file:
                logger.debug(
                    "Adding file-based resource %s to api %s",
                    res_file.path,
                    api_url,
                )
                resource_urls["file"] = api_url + res_file.path
                parsers["file"] = _create_request_parser(res_file)
            if (
                len(resource_urls) == 2
                and resource_urls["query"] == resource_urls["file"]
            ):
                api.add_resource(
                    QueryOrFileUploadResource,  # QueryResource,
                    resource_urls["query"],
                    resource_class_kwargs={
                        "model_api_obj": self,
                        "query_parser": parsers["query"],
                        "file_parser": parsers["file"],
                    },
                )
            else:
                for k, res_url in resource_urls.items():
                    if not res_url:
                        continue
                    api.add_resource(
                        QueryOrFileUploadResource,  # QueryResource,
                        res_url,
                        resource_class_kwargs={
                            "model_api_obj": self,
                            "query_parser": parsers["query"]
                            if k == "query"
                            else None,
                            "file_parser": parsers["file"]
                            if k == "file"
                            else None,
                        },
                    )

        if res_with_id:
            logger.debug(
                "Adding url-id resource %s to api %s",
                res_with_id.path,
                api_url,
            )
            uri_param_name = res_with_id.uri_params[-1].name
            resource_url = (
                api_url
                + res_with_id.parent.path
                + "/<string:some_resource_id>"
            )
            parser = _create_request_parser(res_with_id)
            api.add_resource(
                GetByIdResource,
                resource_url,
                resource_class_kwargs={
                    "model_api_obj": self,
                    "parser": parser,
                    "id_name": uri_param_name,
                },
            )

    def predict_using_model(self, args_dict):
        logger.debug("Prediction input %s", dict(args_dict))
        logger.info("Starting prediction")
        args_ordered_dict = OrderedDict(sorted(args_dict.items()))
        inner_model = self.model_wrapper.contents
        predict_args = [
            self.model_config,
            self.datasources,
            self.datasinks,
            inner_model,
            args_ordered_dict,
        ]
        if hasattr(self.model_wrapper, "__graph"):
            with self.model_wrapper.__graph.as_default():
                logger.info("Restored tensorflow model's graph")
                raw_output = self.model_wrapper.predict(*predict_args)
        else:
            raw_output = self.model_wrapper.predict(*predict_args)

        if (
            hasattr(self.model_wrapper, "__ordered_columns")
            and not resource._order_columns_called
        ):
            logger.warning(
                "Model has been trained on ordered columns, but "
                "prediction does not call function order_columns."
            )

        output = resource.to_plain_python_obj(raw_output)
        logger.debug("Prediction output %s", output)
        return output

    @staticmethod
    def _init_datasources(config):
        logger.info("Initializing datasources...")
        dso, dsi = resource.create_data_sources_and_sinks(
            config, tags="predict"
        )
        logger.info(
            "%s datasource(s) initialized: %s", len(dso), list(dso.keys())
        )
        logger.info(
            "%s datasink(s) initialized: %s", len(dsi), list(dsi.keys())
        )
        if config["api"].get("preload_datasources"):
            logger.info("Preloading datasources...")
            for ds in dso.values():
                _ = ds.get_dataframe()

        return dso, dsi

    @staticmethod
    def _load_model(model_store, model_config):
        logger.info("Loading model...")
        model, meta = model_store.load_trained_model(model_config)
        logger.info(
            "Model loaded: {}, version: {}, created {}".format(
                meta["name"], meta["version"], meta["created"]
            )
        )

        return model


_pd_type_lookup = {
    "object": "string",
    "int64": "integer",
    "float64": "number",
    "bool": "boolean",
    "datetime64": "date",  # https://github.com/raml-org/raml-spec/blob/master/versions/raml-08/raml-08.md#date-representations
    "category": "string"  # plus enum: list(series.cat.categories)
    # RAML Types: any, object, array, union via type expression,
    #             one of the following scalar types: number, boolean, string,
    #             date-only, time-only, datetime-only, datetime, file, integer, or nil
}


def generate_raml(
    complete_conf,
    data_source_name=None,
    data_frame=None,
    resource_name="mythings",
):
    from .model_actions import _get_data_sources_and_sinks
    from urllib.parse import quote_plus

    if data_source_name is not None:
        dso, _ = _get_data_sources_and_sinks(
            complete_conf, tags="", cache=True
        )
        df = dso[data_source_name].get_dataframe()
    elif data_frame is not None:
        df = data_frame
    else:
        raise ValueError("Please provide a data_source_name or a data_frame")
    sample = df.sample(1).reset_index(drop=True)
    api_name = complete_conf["api"]["name"]
    api_version = _get_major_api_version(complete_conf)
    url_start = f"http://127.0.0.1:5000/{quote_plus(api_name)}/{api_version}/{resource_name}?"
    url_params = []
    param_hints = """# can be false if optional, then provide a default here or be prepared to deal with missing values in your prediction
        #default: {}  # only makes sense for required: false
        #minimum: 0  # optional, maximum, minLength and others are also possible."""
    output = f"""
#%RAML 0.8
---
title: Put title of your {api_name} API here
baseUri: https://{{host}}/{api_name}/{{version}}
version: {api_version}  # new version only for API-breaking updates
documentation:
  - title: Example section title
    content: |
      Example section contents


/{resource_name}:  # This should be a plural form of what you're predicting, e.g. "/client_activations"
  get:  # We can support 'post' as well if needed (let us know if necessary)
    description: Briefly describe what {resource_name} exactly you get from this API
    queryParameters:  # For all ways to specify parameters see https://github.com/raml-org/raml-spec/blob/master/versions/raml-08/raml-08.md#named-parameters"""
    for col_name in sample.columns:
        series = sample[col_name]
        type_str = str(series.dtype)
        raml_type = _pd_type_lookup[type_str]
        example = repr(series[0])
        url_params += [quote_plus(col_name) + "=" + quote_plus(example)]
        illegal_chars = ":.,[]'\"\\ \n\t"
        quoted_col_name = (
            f"'{col_name}'"
            if any(c in illegal_chars for c in col_name)
            else col_name
        )
        cleaned_col_name = "".join(
            "_" if c in illegal_chars else c for c in col_name
        )
        output += f"""
      {quoted_col_name}:
        displayName: Friendly Name of {cleaned_col_name}
        type: {raml_type}
        description: Description of what {cleaned_col_name} really is
        example: {example}
        required: true  """ + param_hints.format(
            example
        )
        if type_str == "category":
            output += "\n" + str(list(series.cat.categories))
        param_hints = ""

    output += f"""
    responses:
      200:  # OK
        body:
          application/json:
            # This schema is optional but should fit your example prediction result below:
            schema: |
              {{
                "type": "object",
                "$schema": "http://json-schema.org/draft-03/schema",
                "id": "http://jsonschema.net",
                "required": true,
                "properties": {{
                  "prediction": {{
                    "type": "string",
                    "required": true,
                    "enum": ["Virginica", "Versicolor", "Setosa"]
                  }},
                  "probability": {{
                    "type": "number",
                    "required": false
                  }}
                }}
              }}
            # Provide an example of your prediction result:
            example: |
              {{
                "prediction": "Rainbows and Unicorns!"
              }}
  # /{{test_key}}: # This becomes relevant if you want the API user to provide e.g. ids for the model to look up data to predict for
  #   get:
  #     queryParameters:
  #       hello:
  #         description: some demo query parameter in addition to the uri param
  #         type: string
  #         required: true
  #         enum: ['metric', 'imperial']
  #         #default: 42


  # Example URL API call (to copy and paste into browser (e.g. Chrome) to test):
  # {url_start + '&'.join(url_params)}



  #########################################################################
  # Above is printed an example RAML for your data to copy/paste into your
  # RAML file and adapt to fully define your API.
  #
  # What you should do now:
  #   - Paste the above text into a <mymodel>.raml file (from the line '#%RAML 0.8')
  #   - Find and remove your target variable(s) in the RAML and example URL
  #   - Check and correct the resource name (currently /{resource_name})
  #   - Check the examples for sensitive data
  #   - Fill in the titles and descriptions, adapt as needed
  #
  # Official RAML specification: https://github.com/raml-org/raml-spec/blob/master/versions/raml-08/raml-08.md
  #########################################################################
"""
    return output



import logging

from mllaunchpad import ModelInterface, ModelMakerInterface, order_columns
# from sklearn import tree
# from sklearn.metrics import accuracy_score, confusion_matrix


logger = logging.getLogger(__name__)

# Train this example from the command line:
# python -m mllaunchpad -c TEMPLATE_cfg.yml train
#
# Start REST API:
# python -m mllaunchpad -c TEMPLATE_cfg.yml api
#
# Example API call:
# http://127.0.0.1:5000/TEMPLATE/v0/varieties?sepal.length=4.9&sepal.width=2.4&petal.length=3.3&petal.width=1


class MyExampleModelMaker(ModelMakerInterface):
    """Creates a model
    """

    def create_trained_model(
        self, model_conf, data_sources, data_sinks, old_model=None
    ):
        # df_unordered = data_sources["petals"].get_dataframe()

        # Ordering columns is strongly recommended to guarantee consistency between training and API calls:
        # df = order_columns(df_unordered)

        ...
        model = ...

        return model

    def test_trained_model(self, model_conf, data_sources, data_sinks, model):
        # df_unordered = data_sources["petals_test"].get_dataframe()

        # Ordering columns is strongly recommended to guarantee consistency between training and API calls:
        # df = order_columns(df_unordered)

        ...
        metrics = ...

        return metrics


class MyExampleModel(ModelInterface):
    """Uses the created Data Science Model
    """

    def predict(self, model_conf, data_sources, data_sinks, model, args_dict):
        # Ordering columns is strongly recommended to guarantee consistency between training and API calls:
        # df = order_columns(pd.DataFrame(args_dict, index=[0]))

        ...
        output = ...

        return output
